from dataclasses import dataclass
import numpy
import numpy.typing


@dataclass
class DotMeasurement():
	'''
	Representation of a dot measuring oscillating dipoles.

	Parameters
	----------
	v : float
		The voltage measured at the dot.
	r : numpy.ndarray
		The position of the dot.
	f : float
		The measurement frequency.
	'''
	v: float
	r: numpy.ndarray
	f: float

	def __post_init__(self) -> None:
		self.r = numpy.array(self.r)

	def v_for_point(self, pt: numpy.ndarray) -> float:
		'''
		Returns the voltage for an oscillating dipole represented as a phase space point.

		Parameters
		----------
		pt: A length 7 numpy array representing a dipole, as (px, py, pz, sx, sy, sz, w).

		Returns
		----------
		Returns the voltage from that dipole at the point.
		'''
		p = pt[0:3]  # hardcoded here because chances
		s = pt[3:6]  # are we'll only ever work in 3d.
		w = pt[6]

		return (self._alpha(p, s))**2 * self._b(w)

	def mod_factor_for_point(self, pt: numpy.ndarray) -> float:
		'''
		modification factor for cost function.
		'''
		s = pt[3:6]  # are we'll only ever work in 3d.

		diff = self.r - s
		return numpy.linalg.norm(diff)**2

	def _alpha(self, p: numpy.ndarray, s: numpy.ndarray) -> float:
		diff = self.r - s
		return p.dot(diff) / (numpy.linalg.norm(diff)**3)

	def _b(self, w: float) -> float:
		return (1 / numpy.pi) * (w / (self.f**2 + w**2))

	def cost(self, pts: numpy.ndarray) -> float:
		# 7 because dipole in 3d has 7 degrees of freedom.
		pt_length = 7
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return sum(self.v_for_point(pt) for pt in chunked_pts) - self.v

	def cost2(self, pts: numpy.ndarray) -> float:
		# 7 because dipole in 3d has 7 degrees of freedom.
		pt_length = 7
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]

		mod_factor = numpy.prod([self.mod_factor_for_point(pt) for pt in chunked_pts])
		return (sum(self.v_for_point(pt) for pt in chunked_pts) - self.v) * mod_factor

	def simple_cost(self, pts: numpy.ndarray) -> float:
		# for reduced case, a is constant
		pt_length = 2

		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]
		return sum(pt[0] * self._b(pt[1]) for pt in chunked_pts) - self.v

	def jac_pt(self, pt: numpy.ndarray) -> numpy.ndarray:
		p = pt[0:3]  # hardcoded here because chances
		s = pt[3:6]  # are we'll only ever work in 3d.
		w = pt[6]

		diff = self.r - s
		alpha = self._alpha(p, s)
		b = self._b(w)

		p_divs = 2 * alpha * diff / (numpy.linalg.norm(diff)**3) * b

		r_divs = (-p / (numpy.linalg.norm(diff)**3) + 3 * p.dot(diff) * diff / (numpy.linalg.norm(diff)**5)) * 2 * alpha * b

		f2 = self.f**2
		w2 = w**2

		w_div = alpha**2 * (1 / numpy.pi) * ((f2 - w2) / ((f2 + w2)**2))

		return numpy.concatenate((p_divs, r_divs, w_div), axis=None)

	def jac_pt2(self, pt: numpy.ndarray, cost, mod_factor: float) -> numpy.ndarray:
		p = pt[0:3]  # hardcoded here because chances
		s = pt[3:6]  # are we'll only ever work in 3d.
		w = pt[6]

		diff = self.r - s
		alpha = self._alpha(p, s)
		b = self._b(w)

		p_divs = (2 * alpha * diff / (numpy.linalg.norm(diff)**3) * b) * mod_factor

		s_divs = ((-p / (numpy.linalg.norm(diff)**3) + 3 * p.dot(diff) * diff / (numpy.linalg.norm(diff)**5)) * 2 * alpha * b) * mod_factor
		s_divs_mod = -2 * cost * mod_factor * diff / numpy.linalg.norm(diff)**2

		f2 = self.f**2
		w2 = w**2

		w_div = (alpha**2 * (1 / numpy.pi) * ((f2 - w2) / ((f2 + w2)**2))) * mod_factor

		return numpy.concatenate((p_divs, s_divs + s_divs_mod, w_div), axis=None)

	def simple_jac_pt(self, pt: numpy.ndarray) -> numpy.ndarray:
		a = pt[0]
		w = pt[1]

		f2 = self.f**2
		w2 = w**2

		b = self._b(w)
		w_div = a * (1 / numpy.pi) * ((f2 - w2) / ((f2 + w2)**2))

		return numpy.concatenate((b, w_div), axis=None)

	def jac(self, pts: numpy.ndarray) -> numpy.ndarray:
		# 7 because oscillating dipole in 3d has 7 degrees of freedom.
		pt_length = 7
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]

		return numpy.append([], [self.jac_pt(pt) for pt in chunked_pts])

	def jac2(self, pts: numpy.ndarray) -> numpy.ndarray:
		# 7 because oscillating dipole in 3d has 7 degrees of freedom.
		pt_length = 7
		# creates numpy.ndarrays in groups of pt_length.
		# Will throw problems for irregular points, but that's okay for now.
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]

		cost = self.cost(pts)
		mod_factor = numpy.prod([self.mod_factor_for_point(pt) for pt in chunked_pts])
		return numpy.append([], [self.jac_pt2(pt, cost, mod_factor) for pt in chunked_pts])

	def simple_jac(self, pts: numpy.ndarray) -> numpy.ndarray:
		pt_length = 2
		chunked_pts = [pts[i: i + pt_length] for i in range(0, len(pts), pt_length)]

		return numpy.append([], [self.simple_jac_pt(pt) for pt in chunked_pts])
