from typing import Callable, Sequence
import numpy
import scipy.optimize
from pathfinder.model.oscillating.dot import DotMeasurement
import pathfinder.model.oscillating.util
import logging


class DotOscillatingDipoleModel():
	'''
	Model of n static oscillating with a collection of voltage measurements
	at dots at different positions and frequencies.

	Parameters
	----------
	dots : Sequence[DotMeasurement]
		A collection of dots representing a series of measured noise spectrum values.
	n: int
		The number of dipoles to assume.
	'''
	def __init__(self, dots: Sequence[DotMeasurement], n: int) -> None:
		self.dots = dots
		self.m = len(dots)
		self.n = n

	def __repr__(self) -> str:
		return f'DotOscillatingDipoleModel({repr(list(self.dots))}, {self.n})'

	def costs(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def costs_to_return(pt: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.cost(pt) for dot in self.dots])

		return costs_to_return

	def costs2(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def costs_to_return(pt: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.cost2(pt) for dot in self.dots])

		return costs_to_return

	def simple_costs(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def costs_to_return(pt: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.simple_cost(pt) for dot in self.dots])

		return costs_to_return

	def jac(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def jac_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.jac(pts) for dot in self.dots])

		return jac_to_return

	def jac2(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def jac_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.jac2(pts) for dot in self.dots])

		return jac_to_return

	def simple_jac(self) -> Callable[[numpy.ndarray], numpy.ndarray]:
		def jac_to_return(pts: numpy.ndarray) -> numpy.ndarray:
			return numpy.array([dot.simple_jac(pts) for dot in self.dots])

		return jac_to_return

	def sol(self, initial_dipole=(0.1, 0.1, 0.1), initial_position=(.1, .1, .1), initial_frequency=1, use_root=True):
		initial = numpy.tile(numpy.concatenate((initial_dipole, initial_position, initial_frequency), axis=None), self.n)

		result = scipy.optimize.least_squares(self.costs(), initial, jac=self.jac(), ftol=1e-15, gtol=3e-16)
		result.pathfinder_x = pathfinder.model.oscillating.util.normalize_oscillating_dipole_list(result.x)
		return result

	def sol_minim(self, initial_dipole=(0.1, 0.1, 0.1), initial_position=(.1, .1, .1), initial_frequency=1, use_root=True):
		initial = numpy.tile(numpy.concatenate((initial_dipole, initial_position, initial_frequency), axis=None), self.n)

		cached_costs = self.costs()
		cached_jacobian = self.jac()
		
		def summed(pt):
			current_cost = cached_costs(pt)
			curr_jac = cached_jacobian(pt)
			curr_gradient = .5 * numpy.matmul(numpy.transpose(curr_jac), current_cost)
			return (numpy.sum(current_cost**2), curr_gradient)
	
		result = scipy.optimize.minimize(summed, initial, jac=True, tol=1e-16, options={"gtol": 1e-20})
		result.pathfinder_x = pathfinder.model.oscillating.util.normalize_oscillating_dipole_list(result.x)
		return result
	
	def sol_basinhopping(self, initial_dipole=(0.1, 0.1, 0.1), initial_position=(.1, .1, .1), initial_frequency=1, use_root=True):
		initial = numpy.tile(numpy.concatenate((initial_dipole, initial_position, initial_frequency), axis=None), self.n)

		cached_costs = self.costs()
		cached_jacobian = self.jac()
		
		def summed(pt):
			current_cost = cached_costs(pt)
			curr_jac = cached_jacobian(pt)
			curr_gradient = .5 * numpy.matmul(numpy.transpose(curr_jac), current_cost)
			return (numpy.sum(current_cost**2), curr_gradient)
				
		minimizer_kwargs = {"method": "L-BFGS-B", "jac": True, "tol": 1e-16, "options": {"gtol": 1e-16}}
		
		result = scipy.optimize.basinhopping(summed, initial, minimizer_kwargs=minimizer_kwargs, niter=100)
		result.pathfinder_x = pathfinder.model.oscillating.util.normalize_oscillating_dipole_list(result.x)
		return result

	def sol2(self, initial_dipole=(0.1, 0.1, 0.1), initial_position=(.1, .1, .1), initial_frequency=1, use_root=True):
		initial = numpy.tile(numpy.concatenate((initial_dipole, initial_position, initial_frequency), axis=None), self.n)

		result = scipy.optimize.least_squares(self.costs2(), initial, jac=self.jac2(), ftol=1e-15, gtol=3e-16)
		result.pathfinder_x = pathfinder.model.oscillating.util.normalize_oscillating_dipole_list(result.x)
		return result

	def simple_sol(self, initial_a=0.1, initial_frequency=1):
		initial = numpy.tile(numpy.concatenate((initial_a, initial_frequency), axis=None), self.n)

		result = scipy.optimize.least_squares(self.simple_costs(), initial, jac=self.simple_jac(), ftol=1e-15, gtol=3e-16)
		result.pathfinder_x = result.x
		return result
