import itertools
import logging
import multiprocessing
import numpy
import pathfinder.model.oscillating


def print_result(msg, result):
	logging.info(msg)
	logging.info(f"\tResult: {result.pathfinder_x}")
	logging.info(f"\tSuccess: {result.success}. {result.message}")
	try:
		logging.info(f"\tFunc evals: {result.nfev}")
	except AttributeError:
		pass
	try:
		logging.info(f"\tJacb evals: {result.njev}")
	except AttributeError:
		pass


def try_initial_position(model, expected_result, initial_pos):
	res = model.sol(initial_position=initial_pos)
	return (initial_pos, all(numpy.isclose(res.pathfinder_x, expected_result)))


def main():
	logging.info("Running script...")
	dot_inputs = list(itertools.chain.from_iterable(
		(([0 + o, 0, .01], f), ([0 + o, -1, 0], f), ([-1 + o, 0, -.01], f), ([-1 + o, -1, .01], f)) for f in numpy.arange(1, 10, .5) for o in (0, 0.01)
	))

	dipole = pathfinder.model.oscillating.OscillatingDipole([1, 2, 3], [0, 0, -1], 7)
	expected_result = numpy.array([1, 2, 3, 0, 4, -1, 7])
	dipole_arrangement = pathfinder.model.oscillating.OscillatingDipoleArrangement([dipole])
	dot_measurements = dipole_arrangement.get_dot_measurements(dot_inputs)

	model = pathfinder.model.oscillating.DotOscillatingDipoleModel(dot_measurements, 1)
	logging.info("Finished setting up model")

	work_good = []
	no_work_good = []
	rb = -2
	ru = 3
	points_to_try = [(model, expected_result, (0.01 + dx, 0.01 + dy, 0.01 + dz)) for dx in range(rb, ru, 2) for dy in range(rb, ru, 2) for dz in range(rb, ru, 2)]
	logging.info(f"Will have {len(points_to_try)} points to try")
	logging.info("creating pool...")
	with multiprocessing.Pool() as pool:
		results = pool.starmap(try_initial_position, points_to_try)

		work_good = [x for x,good in results if good]
		no_work_good = [x for x,good in results if not good]

	logging.info(f"work good: [{work_good}]")
	logging.info(f"no work good: [{no_work_good}]")

if __name__ == "__main__":
	logging.basicConfig(level=logging.DEBUG)
	main()
